# -*- coding: utf-8 -*-

from billetterie.utils.mkarchive import TarArchive, ZipArchive
from guichet.models import Registration, Option, Info_Sup
from django.conf import settings

from io import BytesIO
from xlsxwriter import Workbook
import pytz
import os

__all__ = ["Event_export"]

Title_format = None


class Basic_cell(object):
    def __init__(self, value, format=None):
        self.value = value
        self.format = format

    def write(self, ws, row, col):
        ws.write(row, col, self.value, self.format)

    def size(self):
        return len(self.value)


class String_cell(Basic_cell):
    # Use this object to force string
    def write(self, ws, row, col):
        ws.write_string(row, col, self.value, self.format)


class Number_cell(Basic_cell):
    def write(self, ws, row, col):
        ws.write_number(row, col, self.value, self.format)

    def size(self):
        return len(str(self.value))


class Blank_cell(Basic_cell):
    def __init__(self, value=None, format=None):
        self.value = value
        self.format = format

    def write(self, ws, row, col):
        ws.write_blank(row, col, None, self.format)

    def size(self):
        return 0


class Bool_cell(Basic_cell):
    def write(self, ws, row, col):
        ws.write_string(row, col, "oui" if self.value else "non", self.format)

    def size(self):
        return 3


class Date_cell(Basic_cell):
    local_time = pytz.timezone("Europe/Paris")

    def write(self, ws, row, col):
        if not self.value:
            ws.write_blank(row, col, None, self.format)
        else:
            txt = self.value.astimezone(Date_cell.local_time).strftime("%Y-%m-%d %H:%M")
            ws.write_string(row, col, txt, self.format)

    def size(self):
        return 16


class URL_cell(Basic_cell):
    def write(self, ws, row, col):
        ws.write_url(row, col, self.value, self.format)


class IteratorOp(object):
    def __init__(self, formattitle):
        self.titles = []
        self.ops = []
        self.formattitle = formattitle

    def add(self, title, op, formattitle=False):
        if formattitle is False:
            formattitle = self.formattitle
        self.titles.append(String_cell(title, formattitle))
        self.ops.append(op)


class SafeLambda(object):
    def __init__(self, slambda, *opts):
        self.opts = opts
        self.slambda = slambda

    def __call__(self, reg):
        return self.slambda(reg, *self.opts)


class Event_export(object):
    def __init__(self, event):
        self.event = event
        self.archive = None

    def get_zip_archive(self, request):
        with ZipArchive() as f:
            return self._get_archive(request, f)

    def get_tar_archive(self, request):
        with TarArchive() as f:
            return self._get_archive(request, f)

    def _get_archive(self, request, archive):
        self.archive = archive
        archive.addContent(self.get_xlsx().getvalue(), "{}.xlsx".format(self.event.short_url), bytemode=True)
        self.archive = None
        return archive.get_response(request, self.event.short_url)

    def get_xlsx(self):
        f = BytesIO()
        wb = Workbook(f, {"strings_to_numbers": False, "strings_to_formulas": False, "strings_to_urls": False})
        self.titleformat = wb.add_format(
            {
                "bold": True,
            }
        )
        self._get_all_ticket_page(wb.add_worksheet("Tous Billet"))
        for ticket in self.event.tickets.all():
            t_name = ticket.name.translate({ord(i): "_" for i in "[]:*?/\\"})
            name = "{} (n°{})".format(t_name, ticket.id)
            if len(name) > 31:
                postfix = " (n°{})".format(ticket.id)
                l_name = 31 - len(postfix)
                name = t_name[:l_name] + postfix
            self._get_ticket_page(wb.add_worksheet(name), ticket)
        wb.close()

        return f

    def _get_all_ticket_page(self, ws):
        regs = Registration.objects.filter(ticket__event=self.event)
        op = self._get_data(ws)
        self._writesheet(ws, op, regs)

    def _get_ticket_page(self, ws, ticket):
        regs = Registration.objects.filter(ticket=ticket)
        op = self._get_data(ws, ticket)
        self._writesheet(ws, op, regs)

    def _get_data(self, ws, ticket=None):
        op = IteratorOp(self.titleformat)

        op.add("ID", lambda reg: Number_cell(reg.id))
        op.add("Client", lambda reg: String_cell(reg.get_real_payee()))
        if ticket is None:
            op.add("Ticket", lambda reg: String_cell(reg.ticket.name))
        op.add("Date de réservation", lambda reg: Date_cell(reg.purchase_date))
        if Info_Sup.objects.filter(ticket__event=self.event).exists():
            op.add("Etat informations", lambda reg: String_cell(reg.get_info_state_display()))
        op.add("Prix", lambda reg: Number_cell(reg.get_real_amount()))
        op.add("Payé", lambda reg: Bool_cell(reg.paid))
        op.add("Date de paiement", lambda reg: Date_cell(reg.payment_date) if reg.paid else Blank_cell())
        op.add(
            "Moyen de paiement", lambda reg: String_cell(reg.get_payment_method_display()) if reg.paid else Blank_cell()
        )
        op.add("En opposition", lambda reg: Bool_cell(reg.opposed))
        op.add(
            "Raison de mise en opposition",
            lambda reg: String_cell(reg.get_reason_opposed_display()) if reg.opposed else Blank_cell(),
        )
        op.add("Utilisé", lambda reg: Bool_cell(reg.used))
        op.add("Date d'utilisation", lambda reg: Date_cell(reg.used_date) if reg.used else Blank_cell())

        if ticket:
            # options seule
            for opt in Option.objects.filter(ticket=ticket, group=None):
                op.add(
                    "Option {}".format(opt.name), SafeLambda((lambda reg, o: Bool_cell(o in reg.options.all())), opt)
                )

            # group option
            opt_group_list = set()
            for opt in Option.objects.filter(ticket=ticket).exclude(group=None).order_by("group"):
                if opt.group in opt_group_list:
                    continue
                opt_group_list.add(opt.group)
                opts = Option.objects.filter(ticket=ticket, group=opt.group)
                op.add(
                    "Option {}".format(opt.group_name),
                    SafeLambda(
                        (
                            lambda reg, o: String_cell(o.get(options_registration=reg).name)
                            if o.filter(options_registration=reg).exists()
                            else Blank_cell()
                        ),
                        opts,
                    ),
                )

            # Information
            for info in Info_Sup.objects.filter(ticket=ticket):
                op.add(
                    "Statut Info {}".format(info.name),
                    SafeLambda(
                        (
                            lambda reg, i: String_cell(i.received.get(registration=reg).get_state_display())
                            if i.received.filter(registration=reg).exists()
                            else String_cell("Manquante")
                        ),
                        info,
                    ),
                )
                if info.type_info in [Info_Sup.INFO_SHORT_TEXT, Info_Sup.INFO_LONG_TEXT]:
                    op.add(
                        "Info {}".format(info.name),
                        SafeLambda(
                            (
                                lambda reg, i: String_cell(i.received.get(registration=reg).value)
                                if (
                                    i.received.filter(registration=reg).exists()
                                    and i.received.get(registration=reg).value
                                )
                                else Blank_cell()
                            ),
                            info,
                        ),
                    )
                elif info.type_info == Info_Sup.INFO_CHOICE:
                    op.add(
                        "Info {}".format(info.name),
                        SafeLambda(
                            (
                                lambda reg, i: String_cell(i.received.get(registration=reg).value.name)
                                if (
                                    i.received.filter(registration=reg).exists()
                                    and i.received.get(registration=reg).value
                                )
                                else Blank_cell()
                            ),
                            info,
                        ),
                    )
                elif info.type_info in [Info_Sup.INFO_INT, Info_Sup.INFO_FLOAT]:
                    op.add(
                        "Info {}".format(info.name),
                        SafeLambda(
                            (
                                lambda reg, i: Number_cell(i.received.get(registration=reg).value)
                                if (
                                    i.received.filter(registration=reg).exists()
                                    and i.received.get(registration=reg).value is not None
                                )
                                else Blank_cell()
                            ),
                            info,
                        ),
                    )
                elif info.type_info == Info_Sup.INFO_FILE:
                    op.add(
                        "URL Info {}".format(info.name),
                        SafeLambda(
                            (
                                lambda reg, i: URL_cell(settings.BASE_URL + i.received.get(registration=reg).value.url)
                                if (
                                    i.received.filter(registration=reg).exists()
                                    and i.received.get(registration=reg).value
                                )
                                else Blank_cell()
                            ),
                            info,
                        ),
                    )
                    op.add(
                        "Path Info {}".format(info.name),
                        SafeLambda(
                            (
                                lambda reg, i: String_cell(self._add_file_archive(i.received.get(registration=reg)))
                                if (
                                    i.received.filter(registration=reg).exists()
                                    and i.received.get(registration=reg).value
                                )
                                else Blank_cell()
                            ),
                            info,
                        ),
                    )

        return op

    def _add_file_archive(self, info):
        path = self._get_file_path(info)
        if self.archive:
            self.archive.addFile(info.value.path, path)
        return path

    def _get_file_path(self, info):
        return "billet_{}/{}".format(info.registration.id, os.path.basename(info.value.path))

    def _writesheet(self, ws, ops, regs):
        size_col = []
        for col_index, cell in enumerate(ops.titles):
            cell.write(ws, 0, col_index)
            size_col.append(cell.size() + 3)

        for row_index, reg in enumerate(regs):
            for col_index, op in enumerate(ops.ops):
                cell = op(reg)
                cell.write(ws, row_index + 1, col_index)
                size_col[col_index] = max(size_col[col_index], cell.size())

        for col_index, col_width in enumerate(size_col):
            ws.set_column(col_index, col_index, col_width)

        ws.freeze_panes(1, 0)
        ws.autofilter(0, 0, len(regs) - 1, len(ops.ops) - 1)
